"""Base class for machine learning algorithms (classification and regression).
"""

import pandas as pd
import numpy as np
import inspect

# pylint: disable=dangerous-default-value, invalid-name, too-many-instance-attributes, too-many-arguments


class Algorithm:
    """Fit, train and get hyperparams of model.
    Parameters
    ----------
    model : callable
        Model class from library to instantiate
    X_train : pd.DataFrame
        Training features
    X_test : pd.DataFrame
        Test features
    y_train : pd.Series
        Training target values/labels
    y_test : pd.Series
        Test target values/labels
    task : str
        Type of task - 'reg' for regression, 'bin' for binary classification,
        or 'multi' for multiclass classification
    params : dict, optional
        Hyperparameters for model initialization
    proba : float, default=0.5
        Classification threshold for binary problems (ignored for regression)
    """

    def __init__(
        self,
        model,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
        task: str,
        params: dict = {},
        proba: float = 0.5,
    ):
        assert task in ['reg', 'bin', 'multi'], "Task must be one of: 'reg', 'bin', 'multi'"
        
        self.model = model
        self.model_name = type(model()).__name__
        self.task = task
        
        self.X_train = X_train.drop("id_client", axis=1) if "id_client" in X_train.columns else X_train
        self.y_train = y_train
        self.X_test = X_test.drop("id_client", axis=1) if "id_client" in X_test.columns else X_test
        self.y_test = y_test
        
        self.random_state = 2024
        self.estimator = None
        self.y_pred = []
        self.y_pred_val = []
        
        if task != 'reg':
            self.is_binary_class = (task == 'bin')
            self.proba = proba
            self.y_pred_proba = []
        
        self.fit_predict(params)

    def fit(self, params: dict = None):
        """Fits model to X_train, y_train.

        Parameters
        ----------
        params : dict, optional
            Hyperparams for model creation. Defaults to None.
        """
        params = {} if params is None else params
        
        if 'random_state' in inspect.signature(self.model).parameters:
            self.estimator = self.model(**params, random_state=self.random_state)
        else:
            self.estimator = self.model(**params)

        self.estimator.fit(self.X_train, self.y_train.values.ravel())

    def predict(self):
        """Predict values for self.X_test

        Returns
        -------
        np.array
            Predicted values/labels for X_test.
        """
        self.y_pred = self.estimator.predict(self.X_test)
        return self.y_pred

    def predict_val(self, X_val: pd.DataFrame):
        """Predict values for X_val

        Parameters
        ----------
        X_val : pd.DataFrame
            Validation features

        Returns
        -------
        np.array
            Predicted values/labels for X_val.
        """
        self.y_pred_val = self.estimator.predict(X_val)
        return self.y_pred_val

    def fit_predict(self, params={}):
        """Fit model for X_train and predict for X_test.

        Parameters
        ----------
        params : dict, optional
            Params for the model. Defaults to {}.

        Returns
        -------
        np.array
            Predicted values/labels for X_test.
        """
        self.fit(params)
        self.predict()
        return self.y_pred

    def predict_proba(self):
        """
        Predict probabilities for binary classification.
        Only works for binary classification tasks.

        Returns
        -------
        np.array
            Binary predictions based on probability threshold
        """
        assert self.task == 'bin', "predict_proba only works for binary classification tasks"

        self.y_pred_proba = (
            self.estimator.predict_proba(self.X_test)[:, 1] >= self.proba
        ).astype(bool)
        return self.y_pred_proba

    def hyperparams_model(self):
        """Return all hyperparameters of model and feature names.

        Returns
        -------
        dict
            All params.
        """
        params = {
            key: value
            for key, value in self.estimator.get_params().items()
            if value is not None and value is not np.nan
        }
        params["feature_names"] = str(self.X_train.columns.tolist())
        
        if self.task != 'reg':
            params["proba"] = self.proba
            
        return params