"""
Statistics function
"""

from scipy import stats


def test_kruskal_wallis(*groups, variable, dataframe, y="y", alpha=0.02, fill_na=0):
    """
    Statistical Kruskal-Wallis test.

    Parameters
    ----------

    groups : packaging argument, pd.Dataframe
        Dataframe with one of the target groups.
    variable : str
        Column name to analyze
    dataframe : pd.Dataframe
        Dataframe with all target groups
    y : str, optional
        Target column name, by default "y"
    alpha : float, optional
        Statistical significance, by default 0.02
    fill_na : int, optional
        Fill empty rows, by default 0
    """
    krus_group = []
    for group in groups:
        group = group[variable].fillna(fill_na).to_list()
        krus_group.append(group)
    h_statistic, p_value = stats.kruskal(*krus_group)

    print("H-statistic value:", h_statistic)
    print("p-value:", p_value)

    if p_value < alpha:
        print(
            "The null hypothesis is rejected, indicating a statistically significant difference in at least one group pair. See the medians below."
        )
        return dataframe.groupby(y)[variable].median()
    print(
        "There is no basis to reject the null hypothesis—no statistically significant differences exist between the groups."
    )


def test_agosto_pearsona(column, dataframe, alpha=0.02):
    """Test for normality using D'Agostino-Pearson test.

    Parameters
    ----------
    column : str
        Column name to test for normality
    dataframe : pd.DataFrame
        Input dataframe containing the column
    alpha : float, optional
        Statistical significance level, by default 0.02
    """

    stat, p_value = stats.normaltest(dataframe[column])

    print(f"D'Agostino-Pearson statistic value: {stat}")
    print(f"p-value: {p_value}")

    alpha = 0.05
    if p_value > alpha:
        print(
            "There is no basis to reject the null hypothesis: the data follow a normal distribution."
        )
    else:
        print("The null hypothesis is rejected: the data deviate from a normal distribution.")

