"""
Model class
"""

import pandas as pd
import numpy as np

from .algorithm import Algorithm
from .metrics import Metrics
from .evaluate import ModelEvaluator, ShapExplainer
from .experiments import VertexExperiment
from .hyperopt import Hyperopt

# pylint: disable=attribute-defined-outside-init
# pylint: disable=raise-missing-from
# pylint: disable=invalid-name
# pylint: disable=dangerous-default-value
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-instance-attributes


class Modeling:
    """Modeling, metrics and vertex logging.
    Parameters
    ----------
    model : object
        Model class (scikit-learn compatible) that implements fit/predict methods
    X_train : pd.DataFrame
        Training features
    X_test : pd.DataFrame
        Test features
    y_train : pd.DataFrame
        Training labels/values
    y_test : pd.DataFrame
        Test labels/values
    task : str
        Task type: 'reg' for regression, 'bin' for binary classification,
        or 'multi' for multiclass classification
    main_metric : str
        Metric name to optimize during hyperparameter tuning
    model_params : dict, optional
        Initial hyperparameters for model initialization
    proba : float, default=0.5
        Classification threshold for binary problems. Ignored for regression.
    own_metrics : dict, optional
        Custom metrics as {'name': callable(y_true, y_pred)}
    metrics_average : str, default='binary'
        Averaging method for precision/recall in multiclass case. Ignored for regression.
    beta : float, default=2
        Beta value for fbeta_score calculation. Ignored for regression.
    fbeta_average : str, default='binary'
        Averaging method for fbeta_score in multiclass case. Ignored for regression.
    fbeta_weights : list, default=[0.5, 0.5]
        Class weights for fbeta_score calculation. Ignored for regression.
    name_experiment : str, optional
        Name of the experiment in Vertex AI
    data_path : str, optional
        Path to data files in DVC
    labels : list, optional
        Class labels for confusion matrix. Ignored for regression.
    shap_sample_size : int, optional
        Percentage of samples to use for SHAP explanation.
    project : str, optional
        Vertex AI project name
    location : str, optional
        Vertex AI location
    """

    def __init__(
        self,
        model,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
        task: str,
        main_metric: str,
        model_params: dict = None,
        proba: float = 0.5,
        own_metrics: dict = None,
        metrics_average: str = "binary",
        beta: float = 2,
        fbeta_average: str = "binary",
        fbeta_weights: list = [0.5, 0.5],
        name_experiment: str = None,
        data_path: str = None,
        labels: list = None,
        shap_sample_size: float = None,
        project: str = "sts-notebooks",
        location: str = "europe-west4",
    ):
        if task not in ['reg', 'bin', 'multi']:
            raise ValueError("task must be one of: 'reg', 'bin', 'multi'")

        self.model = Algorithm(
            model=model,
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
            task=task,
            params=model_params or {},
            proba=proba if task != 'reg' else None,
        )

        self.metricser = Metrics(
            self.model,
            main_metric=main_metric,
            task=task,
            metrics_average=metrics_average,
            beta=beta,
            fbeta_average=fbeta_average,
            fbeta_weights=fbeta_weights,
            own_metrics=own_metrics,
        )

        # Determine labels and confusion matrix based on task
        if task != 'reg':
            if labels is None:
                labels = [str(label) for label in sorted(self.model.y_test.iloc[:, 0].unique())]
            self.evaluator = ModelEvaluator(self.model, labels)
            confusion_matrix = self.evaluator.get_confusion_matrix()
        else:
            labels = None
            self.evaluator = ModelEvaluator(self.model)  # No labels for regression
            confusion_matrix = None

        self.name_experiment = name_experiment
        self.data_path = data_path
        self.project = project
        self.location = location
        self.task = task

        if name_experiment:
            self.log_vertex_experiments(confusion_matrix, labels) # Pass confusion_matrix and labels

        if shap_sample_size is not None:
            self.shap_explainer = ShapExplainer(
            self.model,
            sample_size=shap_sample_size
            )

    def log_vertex_experiments(self, confusion_matrix, labels):
        """Log results of model to Vertex Experiments."""
        if not self.metricser.metrics:
            self.metricser.calculate_metrics()

        vertex_experiment = VertexExperiment(
            self.name_experiment,
            self.model.model_name,
            self.model.task,
            self.metricser.metrics,
            self.model.hyperparams_model(),
            self.data_path,
            project=self.project,
            location=self.location,
            confusion_matrix=confusion_matrix,  # Use the passed-in value
            labels=labels,  # Use the passed-in value
        )
        vertex_experiment.log_experiment_results_to_vertex()
        self.get_general_metrics()

    def calculate_hyperopt_best_params(
        self, space, n_startup_jobs, hyperopt_iter, is_loss_function=False
    ):
        """Calculate optimal hyperparameters using hyperopt.

        Parameters
        ----------
        space : dict
            Hyperparameter search space definition
        n_startup_jobs : int
            Number of random initialization trials
        hyperopt_iter : int
            Total number of optimization iterations
        is_loss_function : bool, default=False
            If True, metric should be minimized instead of maximized
            
        Returns
        -------
        dict
            Dictionary containing the best parameters found
        """
        self.hyperopt = Hyperopt(self.model, self.metricser, is_loss_function=is_loss_function)
        return self.hyperopt.calculate_hyperopt_best_params(space, n_startup_jobs, hyperopt_iter)

    def get_general_metrics(self):
        """Get all metrics and visualization."""
        print(self.metricser.get_main_metric())
        print(self.metricser.metrics)
        self.evaluator.plot_evaluation()

    def get_metrics_for_val(self, X_val, y_val):
        """Calculate and display model metrics for a validation dataset.
    
        Parameters
        ----------
        X_val : pd.DataFrame
            Validation features
        y_val : pd.Series
            Validation labels/values
        """
        X_test_orig = self.model.X_test
        y_test_orig = self.model.y_test

        self.model.X_test = X_val
        self.model.y_test = y_val

        self.model.predict()
        self.metricser.calculate_metrics()

        self.get_general_metrics()

        self.model.X_test = X_test_orig
        self.model.y_test = y_test_orig
